import logging
log = logging.getLogger(__name__)

import functools

import zarr

from .signal import Signal


# Max size of LRU cache
MAXSIZE = 1024


class ZarrSignal(Signal):

    @classmethod
    def from_zip(cls, path, name):
        array = zarr.open(store=str(path), mode='r', path=f'{name}.zarr')
        return cls(array)

    @classmethod
    def from_path(cls, path):
        path = path.with_suffix('.zarr')
        array = zarr.open(store=str(path), mode='r')
        return cls(array)

    def __init__(self, array):
        self.array = array

    @property
    def fs(self):
        return self.array.attrs['fs']

    @property
    def duration(self):
        return self.array.shape[-1]/self.fs

    def __getitem__(self, slice):
        return self.array[slice]

    @property
    def shape(self):
        return self.array.shape



import logging
log = logging.getLogger(__name__)

import functools

import numpy as np
import pandas as pd
from scipy import signal

from psiaudio import calibration


def get_unique_columns(df, exclude=None):
    return [c for c in df if (len(df[c].unique()) > 1) and (c not in exclude)]


def find_object(node, obj_id):
    if isinstance(obj_id, str):
        obj_id = int(obj_id.split('::')[1])
    if hasattr(node, 'items'):
        if node.get('__id__') == obj_id:
            return node
        for k, v in node.items():
            try:
                return find_object(v, obj_id)
            except KeyError:
                continue
    if isinstance(node, list):
        for i in node:
            try:
                return find_object(i, obj_id)
            except KeyError:
                continue
    raise KeyError('No such node')


class Signal:


    def get_calibration(self):
        cal = self.array.attrs['source']['calibration']
        if isinstance(cal, str):
            cal = find_object(self.array.attrs, cal)
        freq = cal['frequency']
        sens = cal['sensitivity']
        return calibration.InterpCalibration(freq, sens)

    def get_epochs(self, md, offset, duration, detrend=None, downsample=None,
                   columns='auto', cb=None):
        fn = self.get_segments
        return self._get_epochs(fn, md, offset, duration, detrend=detrend,
                                downsample=downsample, columns=columns, cb=cb)

    def get_epochs_filtered(self, md, offset, duration, filter_lb, filter_ub,
                            filter_order=1, detrend='constant',
                            pad_duration=10e-3, downsample=None,
                            columns='auto', cb=None):
        fn = self.get_segments_filtered
        return self._get_epochs(fn, md=md, offset=offset, duration=duration,
                                filter_lb=filter_lb, filter_ub=filter_ub,
                                filter_order=filter_order, detrend=detrend,
                                pad_duration=pad_duration,
                                downsample=downsample, columns=columns, cb=cb)

    def _get_epochs(self, fn, md, *args, columns='auto', **kwargs):
        if columns == 'auto':
            columns = get_unique_columns(md, exclude=['t0', 'key'])
        t0 = md['t0'].values
        arrays = [md[c] for c in columns]
        arrays.append(t0)
        df = fn(t0, *args, **kwargs)
        df.index = pd.MultiIndex.from_arrays(arrays, names=columns + ['t0'])
        return df

    def get_segments(self, times, offset, duration, channel=0, detrend=None,
                     downsample=None, cb=None, cb_n=1000, allow_partial=False):
        if cb is None:
            cb = lambda *a, **kw: None

        times = np.asarray(times)
        indices = np.round((times + offset) * self.fs).astype('i')
        samples = round(duration * self.fs)

        if not allow_partial:
            m = (indices >= 0) & ((indices + samples) < self.shape[-1])
            if not m.all():
                i = np.flatnonzero(~m)
                missing = ', '.join(str(e) for e in i)
                log.warn('Missing epochs %s', missing)
            indices = indices[m]
            index = pd.Index(times[m], name='t0')
        else:
            index = pd.Index(times, name='t0')

        values = []
        n = len(indices)
        for j, i in enumerate(indices):
            # This hack is in-place to handle legacy data that was stored in 1D
            # format rather than the newer 2D format.
            if len(self.shape) == 1:
                if channel != 0:
                    raise ValueError(f'Data is 1D. Cannot load channel {channel}.')
                v = self[i:i+samples]
            else:
                v = self[channel, i:i+samples]
            pad_n = samples - v.shape[-1]
            if pad_n:
                if v.ndim == 2:
                    padding = [(0, 0), (0, pad_n)]
                elif v.ndim == 1:
                    padding = (0, pad_n)
                else:
                    raise ValueError('Unsupported dimensionality for signal')
                v = np.pad(v, padding, mode='constant', constant_values=np.nan)
            values.append(v[np.newaxis])
            if ((j+1) % cb_n) == 0:
                cb((j+1)/n)

        cb(1)

        # We need to ensure that data is cast to double since there are some
        # rare edge-cases in which precision is lost when filtering and
        # downsampling.
        values = np.concatenate(values, axis=0).astype('double')

        if detrend is not None:
            values = signal.detrend(values, axis=-1, type=detrend)

        if downsample is not None:
            values = signal.decimate(values, downsample, axis=-1)
            fs = self.fs / downsample
            samples = values.shape[-1]
        else:
            fs = self.fs

        t = np.arange(samples)/fs + offset
        columns = pd.Index(t, name='time')
        df = pd.DataFrame(values, index=index, columns=columns)
        return df.reindex(times)

    def get_segment(self, time, *args, **kwargs):
        return self.get_segments([time], *args, **kwargs).iloc[0]

    def _get_segments_filtered(self, fn, offset, duration, filter_lb,
                               filter_ub, filter_order=1, detrend='constant',
                               pad_duration=10e-3, downsample=None, cb=None):

        fs = self.fs if downsample is None else self.fs / downsample
        Wn = (filter_lb/(0.5*fs), filter_ub/(0.5*fs))
        b, a = signal.iirfilter(filter_order, Wn, btype='band', ftype='butter')
        df = fn(offset-pad_duration, duration+pad_duration, detrend=detrend,
                downsample=downsample, cb=cb)

        # Attempting to write values *back* into the original df (e.g., via
        # df[:] = result) can take up quite a bit of memory for some bizzare
        # reason (possibly a bug in a specific version of Pandas). To get
        # around this limitation, we should just create a new dataframe with
        # the same index and columns.
        filt = signal.filtfilt(b, a, df.values, axis=-1)
        df_filt = pd.DataFrame(filt, index=df.index, columns=df.columns)
        return df_filt.loc[:, offset:offset+duration]

    def get_random_segments(self, n, offset, duration, detrend, downsample):
        t_min = -offset
        t_max = self.duration-duration-offset
        times = np.random.uniform(t_min, t_max, size=n)
        return self.get_segments(times, offset, duration, detrend,
                                 downsample=downsample)

    def get_segments_filtered(self, times, *args, **kwargs):
        fn = functools.partial(self.get_segments, times)
        return self._get_segments_filtered(fn, *args, **kwargs)

    def get_random_segments_filtered(self, n, *args, **kwargs):
        fn = functools.partial(self.get_random_segments, n)
        return self._get_segments_filtered(fn, *args, **kwargs)



import blosc
import functools
import json
import struct
import numpy as np

from . import Signal


BLOSCPACK_HEADER_LENGTH = 16
BLOSC_HEADER_LENGTH = 16


def decode_uint32(x):
    return struct.unpack('<I', x)[0]


def decode_blosc_header(buffer):
    """ Read and decode header from compressed Blosc buffer.

    Parameters
    ----------
    buffer_ : string of bytes
        the compressed buffer

    Returns
    -------
    settings : dict
        a dict containing the settings from Blosc

    Notes
    -----

    The Blosc 1.1.3 header is 16 bytes as follows:

    | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | A | B | C | D | E | F |
      ^   ^   ^   ^ |     nbytes    |   blocksize   |    ctbytes    |
      |   |   |   |
      |   |   |   +--typesize
      |   |   +------flags
      |   +----------versionlz
      +--------------version

    The first four are simply bytes, the last three are are each unsigned ints
    (uint32) each occupying 4 bytes. The header is always little-endian.
    'ctbytes' is the length of the buffer including header and nbytes is the
    length of the data when uncompressed.

    """
    return {
        'version': buffer[0],
        'versionlz': buffer[1],
        'flags': buffer[2],
        'typesize': buffer[3],
        'nbytes': decode_uint32(buffer[4:8]),
        'blocksize': decode_uint32(buffer[8:12]),
        'ctbytes': decode_uint32(buffer[12:16]),
   }


def clip_chunk(nchunk, chunklen, start, stop, step):
    """Get the limits of a certain chunk based on its length."""
    startb = start - nchunk * chunklen
    stopb = stop - nchunk * chunklen
    # Check limits
    if (startb >= chunklen) or (stopb <= 0):
        return startb, stopb, 0  # null size
    if startb < 0:
        startb = 0
    if stopb > chunklen:
        stopb = chunklen
    # step corrections
    if step > 1:
        # Just correcting startb is enough
        distance = (nchunk * chunklen + startb) - start
        if distance % step > 0:
            startb += (step - (distance % step))
            if startb > chunklen:
                return startb, stopb, 0  # null size
    # Compute size of the clipped block
    blen = get_len_of_range(startb, stopb, step)
    return startb, stopb, blen


def get_len_of_range(start, stop, step):
    """Get the length of a (start, stop, step) range."""
    if start < stop:
        return ((stop - start - 1) // step + 1)
    return 0


class LegacyBcolzArray:

    def __init__(self, rootdir):
        self.filename = rootdir
        self.attrs_filename = rootdir / '__attrs__'
        self.sizes_filename = rootdir / 'meta' / 'sizes'
        self.storage_filename = rootdir / 'meta' / 'storage'
        self.data_path = rootdir / 'data'
        self.attrs = json.loads(self.attrs_filename.read_text())
        self.sizes = json.loads(self.sizes_filename.read_text())
        self.storage = json.loads(self.storage_filename.read_text())
        self.length = self.sizes['shape'][-1]
        self.dtype = np.dtype(self.storage['dtype'])
        self.nbytes = self.sizes['nbytes']
        self.chunklen = self.storage['chunklen']

    def read_chunk(self, nchunk):
        chunk_path = self.data_path / f'__{nchunk}.blp'
        with chunk_path.open('rb') as fh:
            bloscpack_header = fh.read(BLOSCPACK_HEADER_LENGTH)
            blosc_header_raw = fh.read(BLOSC_HEADER_LENGTH)
            blosc_header = decode_blosc_header(blosc_header_raw)
            ctbytes = blosc_header['ctbytes']
            fh.seek(-BLOSC_HEADER_LENGTH, 1)
            b = fh.read(ctbytes)
        return np.frombuffer(blosc.decompress(b), dtype=self.dtype)

    def __getitem__(self, key):
        start, stop, step = key.indices(self.length)
        if start > stop:
            raise ValueError('Invalid slice')

        n = get_len_of_range(start, stop, step)
        arr = np.empty(shape=(n,), dtype=self.dtype)
        if n == 0:
            return arr

        if self.dtype.char == 'O':
            raise ValueError('Unsupported dtype')

        lb = start // self.chunklen
        ub = stop // self.chunklen + 1

        i = 0
        for n_chunk in range(lb, ub):
            startb, stopb, blen = clip_chunk(n_chunk, self.chunklen, start, stop, step)
            if blen == 0:
                continue
            arr[i:i + blen] = self.read_chunk(n_chunk)[startb:stopb:step]
            i += blen
        return arr


class LegacyBcolzSignal(Signal):

    def __init__(self, base_path):
        self.base_path = base_path

    @property
    @functools.lru_cache()
    def array(self):
        return LegacyBcolzArray(rootdir=self.base_path)

    @property
    def fs(self):
        return self.array.attrs['fs']

    @property
    def duration(self):
        return self.array.length/self.fs

    def __getitem__(self, slice):
        return self.array[slice]

    @property
    def shape(self):
        return (self.array.length,)



import logging
log = logging.getLogger(__name__)

from collections import OrderedDict
import functools
from glob import glob
import json
import shutil
from pathlib import Path
import os.path

import bcolz
import numpy as np
import pandas as pd
from scipy import signal

from . import Signal


# Max size of LRU cache
MAXSIZE = 1024


def repair_carray_size(path):
    chunk_wildcard = os.path.join(path, 'data', '*.blp')
    sizes_filename = os.path.join(path, 'meta', 'sizes')
    storage_filename = os.path.join(path, 'meta', 'storage')

    with open(sizes_filename, 'r') as fh:
        sizes = json.load(fh)

    with open(storage_filename, 'r') as fh:
        storage = json.load(fh)

    if sizes['shape'] != [0]:
        raise ValueError('Data seems fine')

    chunklen = storage['chunklen']
    n_chunks = len(glob(chunk_wildcard))
    sizes['shape'] = [n_chunks * chunklen]

    # Backup the file before overwriting.
    shutil.copy(sizes_filename, sizes_filename + '.old')
    with open(sizes_filename, 'w') as fh:
        json.dump(sizes, fh)


def load_ctable_as_df(path, decode=True, archive=True):
    csv_path = f'{path}.csv'
    if os.path.exists(csv_path):
        return pd.io.parsers.read_csv(csv_path)
    table = bcolz.ctable(rootdir=path)
    df = carray_to_dataframe(table)
    if decode:
        for c in table.cols:
            if table[c].dtype.char == 'S':
                df[c] = df[c].str.decode('utf8')

    if archive:
        df.to_csv(csv_path, index=False)
    return df


def carray_to_dataframe(ctable, columns=None, orient='columns'):
    # Right now only legacy bcolz is here.
    if orient == 'index':
        keys = ctable.names
    else:
        keys = ctable.names if columns is None else columns
        columns = None

    # Use a generator here to minimize the number of column copies
    # existing simultaneously in-memory
    df = pd.DataFrame.from_dict(
        OrderedDict((key, ctable[key][:]) for key in keys),
        columns=columns, orient=orient)
    return df


class BcolzSignal(Signal):

    def __init__(self, base_path):
        self.base_path = base_path

    @property
    @functools.lru_cache()
    def array(self):
        return bcolz.carray(rootdir=self.base_path)

    @property
    def fs(self):
        return self.array.attrs['fs']

    @property
    def duration(self):
        return self.array.shape[-1]/self.fs

    @property
    def ndim(self):
        return self.array.ndim

    def __getitem__(self, slice):
        return self.array[slice]

    @property
    def shape(self):
        return self.array.shape



import pandas as pd

from . import Recording


def _get_unique(df, level_name):
    choices = df.index.unique(level_name)
    if len(choices) != 1:
        choices = ', '.join(str(c) for c in choices)
        m = f'Must specify {level_name}. Options are {choices}.'
        raise ValueError(m)
    return choices[0]


def repair_golay_sens_name(fh):
    if 'golay_sens' in fh.ttable_names:
        return
    for sens_name in ('golay_sensitivity', 'golay_summary', 'sensitivity'):
        if sens_name in fh.ttable_names:
            break
    else:
        raise ValueError(f'{fh} does not contain Golay data')
    old_filename = (fh.base_path / sens_name).with_suffix('.csv')
    new_filename = (fh.base_path / 'golay_sens.csv')
    old_filename.rename(new_filename)
    fh._refresh_names()


def repair_chirp_sens_name(fh):
    if 'chirp_sens' in fh.ttable_names:
        return
    for sens_name in ('chirp_sensitivity', 'chirp_summary', 'sensitivity'):
        if sens_name in fh.ttable_names:
            break
    else:
        raise ValueError(f'{fh} does not contain chirp data')
    old_filename = (fh.base_path / sens_name).with_suffix('.csv')
    new_filename = (fh.base_path / 'chirp_sens.csv')
    old_filename.rename(new_filename)
    fh._refresh_names()


class CalibrationFile(Recording):

    _ttable_indices = {
        'tone_sens': ['channel_name', 'frequency'],
        'golay_sens': ['n_bits', 'output_gain', 'frequency'],
    }

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        if 'golay' in self.base_path.name.lower():
            repair_golay_sens_name(self)
        elif 'chirp' in self.base_path.name.lower():
            repair_chirp_sens_name(self)

    def get_tone_calibration(self, channel_name=None):
        '''
        Return PointCalibration as derived from tone data
        '''
        from psiaudio.calibration import PointCalibration
        if channel_name is None:
            channel_name = _get_unique(self.tone_sens, 'channel_name')
        sens = self.tone_sens.loc[channel_name]['sens']
        return PointCalibration(sens.index, sens.values)

    def _get_golay_data(self, n_bits=None, output_gain=None):
        for epoch_name in ('pt_epoch', 'epoch'):
            if epoch_name in self.carray_names:
                break
        else:
            m = f'Path {self.base_path} does not appear to contain Golay data'
            raise ValueError(m)

        sens_filename = self.base_path / 'golay_sens.csv'
        sensitivity = pd.io.parsers.read_csv(sens_filename)
        if n_bits is None:
            n_bits = sensitivity['n_bits'].max()
        if output_gain is None:
            m = sensitivity['n_bits'] == n_bits
            output_gain = sensitivity.loc[m, 'output_gain'].max()
        m_n_bits = sensitivity['n_bits'] == n_bits
        m_output_gain = sensitivity['output_gain'] == output_gain
        m = m_n_bits & m_output_gain
        mic_freq = sensitivity.loc[m, 'frequency'].values
        mic_sens = sensitivity.loc[m, 'sens'].values
        mic_phase = sensitivity.loc[m, 'phase'].values
        epoch = getattr(self, epoch_name)
        return {
            'attrs': {
                'source': self.base_path,
                'n_bits': n_bits,
                'output_gain': output_gain,
            },
            'frequency': mic_freq,
            'sensitivity': mic_sens,
            'phase': mic_phase,
            'fs': epoch.fs,
        }

    def get_golay_calibration(self, n_bits=None, output_gain=None):
        '''
        Return GolayCalibration as derived from golay data
        '''
        from psi.controller.calibration.api import GolayCalibration
        data = self._get_golay_data(n_bits, output_gain)
        return GolayCalibration(**data)



import logging
log = logging.getLogger(__name__)

import functools
from pathlib import Path
import zipfile

import numpy as np
import pandas as pd


class Recording:
    '''
    Wrapper around a recording created by psiexperiment

    Parameters
    ----------
    base_path : :obj:`str` or :obj:`pathlib.Path`
        Folder containing recordings

    Attributes
    ----------
    base_path : pathlib.Path
        Folder containing recordings
    carray_names : set
        List of Bcolz carrays in this recording
    ctable_names : set
        List of Bcolz ctables in this recording
    ttable_names : set
        List of CSV-formatted tables in this recording

    The `__getattr__` method is implemented to allow accessing arrays and
    tables by name. For example, if you have a ctable called `erp_metadata`:

        recording = Recording(base_path)
        erp_md = recording.erp_metadata

    When using this approach, all tables are loaded into memory and returned as
    instances of `pandas.DataFrame`. All arrays are returned as `Signal`
    instances. Signal instances do not load the data into memory until the data
    is requested.
    '''

    #: Mapping of names for CSV-formatted table to a list of columns that
    #: should be used as indices. For example:
    #:     {'tone_sens': ['channel_name', 'frequency']}
    #: This attribute is typically used by subclasses to automate handling of
    #: loading tables into DataFrames.
    _ttable_indices = {}

    #: This is the name of the table containing the settings that we may wish
    #: to extract.
    _setting_table = None

    def __init__(self, base_path, setting_table=None):
        self._setting_table = setting_table
        self.base_path = Path(base_path)
        if self.base_path.suffix == '.zip':
            self._store = ZipStore(self.base_path, self._ttable_indices)
        elif self.base_path.is_dir():
            self._store = DirStore(self.base_path, self._ttable_indices)
        else:
            raise ValueError(f'Unrecognized recording format at {base_path}')

    def get_setting(self, setting_name):
        '''
        Return value for setting

        Parameters
        ----------
        setting_name : string
            Setting to extract

        Returns
        -------
        object
            Value of setting

        Raises
        ------
        ValueError
            If the setting is not identical across all trials.
        KeyError
            If the setting does not exist.
        '''
        table = getattr(self, self._setting_table)
        values = np.unique(table[setting_name])
        if len(values) != 1:
            raise ValueError('{name} is not unique across all epochs.')
        return values[0]

    def get_setting_default(self, setting_name, default):
        '''
        Return value for setting

        Parameters
        ----------
        setting_name : string
            Setting to extract
        default : obj
            Value to return if setting doesn't exist.

        Returns
        -------
        object
            Value of setting

        Raises
        ------
        ValueError
            If the setting is not identical across all trials.
        '''
        try:
            return self.get_setting(setting_name)
        except KeyError:
            return default

    def __getattr__(self, attr):
        return self._store.__getattr__(attr)

    def __repr__(self):
        lines = [f'Recording at {self.base_path.name} with:']
        if self._store.zarr_names:
            lines.append(f'* Zarr arrays {self._store.zarr_names}')
        if self._store.carray_names:
            lines.append(f'* Bcolz carrays {self._store.carray_names}')
        if self._store.ctable_names:
            lines.append(f'* Bcolz ctables {self._store.ctable_names}')
        if self._store.ttable_names:
            lines.append(f'* CSV tables {self._store.ttable_names}')
        return '\n'.join(lines)


class BaseStore:

    def _refresh_names(self):
        '''
        Utility function to refresh list of signals and tables
        '''
        raise NotImplementedError

    def __getattr__(self, attr):
        if attr in self.zarr_names:
            return self._load_zarr_signal(attr)
        if attr in self.carray_names:
            return self._load_bcolz_signal(attr)
        if attr in self.ctable_names:
            return self._load_bcolz_table(attr)
        if attr in self.ttable_names:
            return self._load_text_table(attr)
        raise AttributeError(attr)

    @functools.lru_cache()
    def _load_bcolz_signal(self, name):
        from .bcolz_tools import BcolzSignal
        return BcolzSignal(self.base_path / name)

    @functools.lru_cache()
    def _load_zarr_signal(self, name):
        from .zarr_tools import ZarrSignal
        return ZarrSignal.from_path(self.base_path / name)

    @functools.lru_cache()
    def _load_bcolz_table(self, name):
        from .bcolz_tools import load_ctable_as_df
        return load_ctable_as_df(self.base_path / name)

    def _get_text_table_stream(self, name):
        raise NotImplementedError

    @functools.lru_cache()
    def _load_text_table(self, name):
        import pandas as pd
        index_col = self._ttable_indices.get(name, None)
        with self._get_text_table_stream(name) as stream:
            df = pd.read_csv(stream, index_col=index_col)

        drop = [c for c in df.columns if c.startswith('Unnamed:')]
        return df.drop(columns=drop)


class DirStore(BaseStore):

    def __init__(self, base_path, ttable_indices):
        self.base_path = Path(base_path)
        self._ttable_indices = ttable_indices
        self._refresh_names()

    def _get_text_table_stream(self, name):
        path = (self.base_path / name).with_suffix('.csv')
        return path.open()

    def _refresh_names(self):
        bp = self.base_path
        self.carray_names = {d.parent.stem for d in bp.glob('*/meta')}
        self.ctable_names = {d.parent.parent.stem for d in bp.glob('*/*/meta')}
        self.ttable_names = {d.stem for d in bp.glob('*.csv')}
        self.zarr_names = {d.stem for d in bp.glob('*.zarr')}


class ZipStore(BaseStore):

    def __init__(self, base_path, ttable_indices):
        self.base_path = Path(base_path)
        self._ttable_indices = ttable_indices
        self.zip_fh = zipfile.ZipFile(base_path)
        self._refresh_names()

    def _refresh_names(self):
        self.carray_names = set()
        self.ctable_names = set()
        self.ttable_names = set()
        self.zarr_names = set()
        for name in self.zip_fh.namelist():
            if name.endswith('.zarr/'):
                self.zarr_names.add(name.split('.', 1)[0])
            elif name.endswith('.csv'):
                self.ttable_names.add(name.split('.', 1)[0])
            elif name.endswith('meta'):
                raise ValueError('ZipRecording does not support bcolz')

    def _get_text_table_stream(self, name):
        return self.zip_fh.open(f'{name}.csv')

    @functools.lru_cache()
    def _load_zarr_signal(self, name):
        from .zarr_tools import ZarrSignal
        return ZarrSignal.from_zip(self.base_path, name)
